var url = require('url');

module.exports = {
  parse: function parse(str) {
    if( typeof str !== 'string') return;
    var config;
    if(str.charAt(0) === '/') {
      config = str.split(' ');
      return { host: config[0], database: config[1] };
    }
    if(/ |%[^a-f0-9]|%[a-f0-9][^a-f0-9]/i.test(str)) {
      str = encodeURI(str).replace(/\%25(\d\d)/g, "%$1");
    }
    var result = url.parse(str, true);
    config = {};

    if (result.query.application_name) {
      config.application_name = result.query.application_name;
    }
    if (result.query.fallback_application_name) {
      config.fallback_application_name = result.query.fallback_application_name;
    }
    config.protocol = null;
    try { config.protocol = result.protocol.replace(/\:\/?\/?$/gmi,"").trim()} catch(e){ }
    config.port = result.port;
    if(result.protocol == 'socket:') {
      config.host = decodeURI(result.pathname);
      config.database = result.query.db;
      config.client_encoding = result.query.encoding;
      return config;
    }
    config.host = result.hostname;

    var pathname = result.pathname;
    if (pathname && pathname.charAt(0) === '/') {
      pathname = result.pathname.slice(1) || null;
    }
    config.database = pathname && decodeURI(pathname);

    var auth = (result.auth || ':').split(':');
    config.user = auth[0];
    config.password = auth.splice(1).join(':');

    var ssl = result.query.ssl;
    if (ssl === 'true' || ssl === '1') {
      config.ssl = true;
    }

    return config;
  }
}
